/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "actiondef.h"
#include "encoding.h"
#include "input.h"

actiondef::actiondef()
{
	names = NULL;
	code = NULL;
	numname = numparam = 0;
	params = NULL;
}

actiondef::~actiondef()
{
	for(uint32_t i=0;i<numname;i++)
		delete names[i];
	free(names);
	delete code;
	if(params)
	{
		for(uint32_t i=0;i<numparam;i++)
			delete params[i];
		free(params);
	}
}

bool actiondef::hasname(const char *name) const
{
	for(uint32_t i=0;i<numname;i++)
		if(!strcmp(names[i],name))
			return true;
	return false;
}

bool actiondef::tryassign(encoding *e) const
{
	for(uint32_t i=0;i<e->numfields;i++)
		e->fields[i].relevant = false;
	for(uint32_t i=0;i<numparam;i++)
	{
		const char *param = params[i];
		while(param)
		{
			const char *terminator = strchr(param,':');
			size_t len;
			if(terminator)
				len = (terminator++)-param;
			else
				len = strlen(param);
			uint32_t j;
			for(j=0;j<e->numfields;j++)
				if((strlen(e->fields[j].name) == len) && !strncmp(param,e->fields[j].name,len))
				{
					e->fields[j].relevant = true;
					break;
				}
			if(j == e->numfields)
				return false;
			param = terminator;
		}
	}
	e->action = this;
	return true;
}

bool actiondef::trymerge(actiondef *a)
{
	/* Keep it simple for now */
	if((numparam != a->numparam) || (nonstandard != a->nonstandard) || strcmp(code,a->code))
		return false;
	/* Check parameters. The order can be different, but they must all be present in each other. */
	for(uint32_t i=0;i<numparam;i++)
	{
		uint32_t j;
		for(j=0;j<numparam;j++)
			if(!strcmp(params[i],a->params[j]))
				break;
		if(j == numparam)
			return false;
	}
	/* Merge them */
	names = (const char **) realloc(names,sizeof(char *)*(numname+a->numname));
	assert(names);
	memcpy(names+numname,a->names,sizeof(char *)*a->numname);
	numname += a->numname;
	a->numname = 0;
	delete a;
	return true;
}

actiondef *actiondef::parse()
{
	/* Syntax is similar to C function definition:
	   name(params)
	   {
	     code
	   }
	   Or:
	   X as if Y
	*/
	string code;
	char buf[1024];
	if(!input_getline(buf,sizeof(buf)))
		return NULL;
	char *s = buf;
	if(!isalpha(*s) && (*s != '_'))
	{
		/* Comment or blank line */
		while(*s && isspace(*s) && (*s != '#'))
			s++;
		if(*s && (*s != '#'))
			input_errorc(s-buf,"Unexpected character '%c' while looking for action name\n",*s);
		return NULL;
	}
	while(isalnum(*s) || (*s == '_'))
		s++;
	if(!strncmp(s," as if ",7))
	{
		/* 'X as if Y' action, i.e. use action Y for encoding X, even if the fields don't match
		   Just search the encoding list for 'X' and assign it action 'Y' (assuming that action exists!) */
		/* First trim s to just the Y part */
		*s = 0;
		s += 7;
		char *s2 = s;
		while(isalnum(*s2) || (*s2 == '_'))
			s2++;
		if((*s2 && !isspace(*s2) && (*s2 != '#')) || (s == s2))
		{
			input_errorc(s2-buf,"Unexpected character '%c' while handling 'as if' action\n",*s2);
			return NULL;
		}
		*s2 = 0;
		/* Now find encoding 'Y' and make sure that it's assigned to action 'Y' */ 
		size_t i,j;
		for(i=0;i<allactions.size();i++)
		{
			if(allactions[i]->hasname(s))
				break;
		}
		if(i == allactions.size())
		{
			input_error("Couldn't find action '%s' for 'as if' action\n",s);
			return NULL;
		}
		for(j=0;j<allencodings.size();j++)
			if(!strcmp(allencodings[j]->name,s))
				break;
		if(j == allencodings.size())
		{
			input_error("Couldn't find encoding '%s' for 'as if' action\n",s);
			return NULL;
		}
		/* Make sure encoding 'Y' uses action 'Y' */
		if(allencodings[j]->action)
		{
			if(allencodings[j]->action != allactions[i])
			{
				input_error("Expected encoding '%s' to be using an action of the same name, but it's currently using action '%s'. Can't assign 'as if' action.\n",s,allencodings[j]->action->names[0]);
				return NULL;
			}
		}
		else
		{
			/* Try assigning them */
			if(!allactions[i]->tryassign(allencodings[j]))
			{
				input_error("Action '%s' is incompatible with encoding of same name; can't use as basis for 'as if' action\n",s);
				return NULL;
			}
		}
		/* Now find all encodings 'X', set them to use allencodings[j]->action, and forcibly set X as a kin of allencodings[j] */
		bool any = false;
		for(i=0;i<allencodings.size();i++)
			if(!strcmp(allencodings[i]->name,buf))
			{
				any = true;
				if(allencodings[i]->action)
				{
					input_error("Can't assign encoding '%s' to action '%s'; already assigned to action '%s'\n",buf,s,allencodings[i]->action->names[0]);
					return NULL;
				}
				allencodings[i]->action = allencodings[j]->action;
				/* Link up the kinship
				   TODO - This is probably wrong in many ways */
				allencodings[i]->kin = allencodings[j];
				allencodings[i]->kin_next = allencodings[j]->kin_next;
				allencodings[j]->kin_next = allencodings[i];
			}
		if(!any)
		{
			input_error("Couldn't find encoding '%s' for 'as if' action\n",buf);
			return NULL;
		}
		return NULL;
	}
	if(*s != '(')
	{
		input_errorc(s-buf,"Expected '(' but found '%c'\n",*s);
		return NULL;
	}
	actiondef *a = new actiondef();
	a->names = (const char **) malloc(sizeof(char *));
	assert(a->names);
	char *temp = new char[(s-buf)+1];
	assert(a && temp);
	strncpy(temp,buf,s-buf);
	temp[s-buf] = 0;
	a->names[0] = temp;
	a->numname = 1;
	a->nonstandard = false;
	s++;
	do {
		while(isspace(*s))
			s++;
		if(a->numparam && (*s == ','))
		{
			do {
				s++;
			} while(isspace(*s));
		}
		if(*s == ')')
			break;
		if(!isalpha(*s) && (*s != '_'))
		{
			input_errorc(s-buf,"Bad character '%c' in parameter list",*s);
			delete a;
			return NULL;
		}
		temp = s;
		while(isalnum(*s) || (*s == '_') || (*s == ':'))
			s++;
		char *temp2 = new char[(s-temp)+1];
		assert(temp2);
		strncpy(temp2,temp,s-temp);
		temp2[s-temp] = 0;
		if(strcmp(temp2,"nonstandard"))
		{
			a->params = (const char **) realloc(a->params,sizeof(char *)*(a->numparam+1));
			assert(a->params);
			a->params[a->numparam++] = temp2;
		}
		else
		{
			delete[] temp2;
			if(a->nonstandard)
			{
				input_error("Action %s has duplicated parameter: nonstandard\n",a->names[0]);
				delete a;
				return NULL;
			}
			a->nonstandard = true;
		}
	} while(1);
	s++;
	/* Verify there aren't any duplicate parameters */
	for(uint32_t i=0;i<a->numparam;i++)
		for(uint32_t j=i+1;j<a->numparam;j++)
			if(!strcmp(a->params[i],a->params[j]))
			{
				input_error("Action %s has duplicated parameter: %s\n",a->names[0],a->params[i]);
				delete a;
				return NULL;
			}
	/* Remainder of line should be blank or comment */
	while(isspace(*s))
		s++;
	if(*s && (*s != '#'))
	{
		input_errorc(s-buf,"Unexpected character '%c' after parameter list\n",*s);
		delete a;
		return NULL;
	}
	if(!input_getline(buf,sizeof(buf)))
	{
		delete a;
		return NULL;
	}
	if(buf[0] != '{')
	{
		input_error("Unrecognised line following action parameter list; should be '{'\n");
		delete a;
		return NULL;
	}
	do {
		if(!input_getline(buf,sizeof(buf)))
		{
			delete a;
			return NULL;
		}
		if(buf[0] == '}')
			break;
		code.append(buf);
	} while(1);
	temp = new char[code.length()+1];
	assert(temp);
	code.copy(temp,code.length());
	temp[code.length()] = 0;
	a->code = temp;
	return a;
}
