/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "common.h"
#include "node.h"
#include "encoding.h"
#include "constraint.h"
#include "decgen.h"
#include "funcdef.h"

encodingset *compute_kin(const encodingset &e)
{
	encodingset *kin = new encodingset;
	kin->reserve(e.size()); /* Reserve worst-case size to avoid reallocs */
	for(size_t i=0;i<e.size();i++)
	{
		encoding *root = (encoding *) (e[i]->kin?e[i]->kin:e[i]);
		size_t j;
		for(j=0;j<kin->size();j++)
			if((*kin)[j] == root)
				break;
		if(j == kin->size())
			kin->push_back(root);
	}
	return kin;
}

static inline bool restrict_recurse_delta(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num,const constraint *cc1,const constraint *cc2)
{
	/* return true iff there is an opcode for which cc1, cc2, and c are all true
	   Ideally these constraints should be rewritten so that cc1 and cc2 are part of 'c', as it would allow cc1 and cc2 to be optimised further */
	uint32_t mask2 = mask | (opcodebits & ~cc1->getfields());
	uint32_t mask2b = mask | cc1->getfields();
	uint32_t mask3 = mask2b | (opcodebits & ~cc2->getfields());
	mask2b |= cc2->getfields();
	uint32_t opcode = 0;
	do {
		if(cc1->eval(opcode | fixed))
		{
			uint32_t opcode2 = 0;
			uint32_t fixed2 = opcode | fixed;
			do {
				if(cc2->eval(opcode2 | fixed2) && restrict_recurse(opcode2 | fixed2,mask2b,c,num))
					return true;
				opcode2++;
				while(opcode2 & mask3)
					opcode2 = (opcode2+mask3)&~mask3;
			} while(opcode2 & opcodebits);
		}
		opcode++;
		while(opcode & mask2)
			opcode = (opcode+mask2)&~mask2;
	} while(opcode & opcodebits);
	return false;
}

static inline bool restrict_recurse_epsilon(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num,const constraint *cc1,const constraint *cc2)
{
	/* return true iff there is an opcode for which !cc1, !cc2, and c are all true
	   Ideally these constraints should be rewritten so that cc1 and cc2 are part of 'c', as it would allow cc1 and cc2 to be optimised further */
	uint32_t mask2 = mask | (opcodebits & ~cc1->getfields());
	uint32_t mask2b = mask | cc1->getfields();
	uint32_t mask3 = mask2b | (opcodebits & ~cc2->getfields());
	mask2b |= cc2->getfields();
	uint32_t opcode = 0;
	do {
		if(!cc1->eval(opcode | fixed))
		{
			uint32_t opcode2 = 0;
			uint32_t fixed2 = opcode | fixed;
			do {
				if(!cc2->eval(opcode2 | fixed2) && restrict_recurse(opcode2 | fixed2,mask2b,c,num))
					return true;
				opcode2++;
				while(opcode2 & mask3)
					opcode2 = (opcode2+mask3)&~mask3;
			} while(opcode2 & opcodebits);
		}
		opcode++;
		while(opcode & mask2)
			opcode = (opcode+mask2)&~mask2;
	} while(opcode & opcodebits);
	return false;
}

static inline bool restrict_recurse_beta(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num,const constraint *cc);

static inline bool restrict_recurse_gamma(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num,const constraint *cc)
{
	/* return true iff there is an opcode for which !cc and c are all true */
	/* The fact that we're dealing with a set of expressions that define a set of values means that we can't directly perform many of De Morgan's laws.
	   However, we can do the following:
	   {lnot(lnot(X))} -> {X} (handled by _beta function)
	   {lnot(land(X,Y))} -> {lor(lnot(X),lnot(Y))} (iterating over X & Y independently)
	   {lnot(lor(X,Y))} -> {land(lnot(X),lnot(Y))} perform two nested iterations for !X and !Y */
	if(cc->gettype() == EFUNC)
	{
		const constraint_func *f = static_cast<const constraint_func *>(cc);
		if(f->getfunc() == funcdef_lnot)
			return restrict_recurse_beta(fixed,mask,c,num,f->getparams()[0]);
		else if(f->getfunc() == funcdef_land)
		{
			return restrict_recurse_gamma(fixed,mask,c,num,f->getparams()[0]) || restrict_recurse_gamma(fixed,mask,c,num,f->getparams()[1]);
		}
		else if(f->getfunc() == funcdef_lor)
		{
			return restrict_recurse_epsilon(fixed,mask,c,num,f->getparams()[0],f->getparams()[1]);
		}
	}
	uint32_t mask2 = mask | (opcodebits & ~cc->getfields());
	uint32_t opcode=0;
	do {
		if(!cc->eval(opcode | fixed) && restrict_recurse(opcode | fixed,mask | cc->getfields(),c,num))
			return true;
		opcode++;
		while(opcode & mask2)
			opcode = (opcode+mask2)&~mask2;
	} while(opcode & opcodebits);
	return false;
}

static inline bool restrict_recurse_beta(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num,const constraint *cc)
{
	/* return true iff there is an opcode for which cc1 and c are all true */
	/* The fact that we're dealing with a set of expressions that define a set of values means that we can't directly perform many of De Morgan's laws.
	   However, we can do the following:
	   {lnot(X)} -> handled by _gamma function
	   {lor(X,Y)} -> {X} or {Y}
	   {land(X,Y)} -> perform two nested iterations for X and Y (can't easily optimise much more, due to potential interdependencies between X & Y or X/Y and another constraint) */
	if(cc->gettype() == EFUNC)
	{
		const constraint_func *f = static_cast<const constraint_func *>(cc);
		if(f->getfunc() == funcdef_lnot)
		{
			return restrict_recurse_gamma(fixed,mask,c,num,f->getparams()[0]);
		}
		else if(f->getfunc() == funcdef_lor)
		{
			return restrict_recurse_beta(fixed,mask,c,num,f->getparams()[0]) || restrict_recurse_beta(fixed,mask,c,num,f->getparams()[1]);
		}
		else if(f->getfunc() == funcdef_land)
		{
			return restrict_recurse_delta(fixed,mask,c,num,f->getparams()[0],f->getparams()[1]);
		}
	}
	/* Can't optimise - just iterate */
	uint32_t mask2 = mask | (opcodebits & ~cc->getfields());
	uint32_t opcode=0;
	do {
		if(cc->eval(opcode | fixed) && restrict_recurse(opcode | fixed,mask | cc->getfields(),c,num))
			return true;
		opcode++;
		while(opcode & mask2)
			opcode = (opcode+mask2)&~mask2;
	} while(opcode & opcodebits);
	return false;
}

bool restrict_recurse(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num)
{
	if(!num)
		return true;
	/* Find a constraint to evaluate
	   We favour ones that are partially (or fully!) restricted over ones that aren't restricted */
	uint32_t i;
	for(i=0;i<num;i++)
		if(c[i]->getfields() & mask)
			break;
	if(i == num)
	{
		/* We didn't find any that are partially restricted; instead, look for one that intersects another */
		for(i=0;i<num;i++)
			for(uint32_t j=i+1;j<num;j++)
				if(c[i]->getfields() & c[j]->getfields())
					goto found;
		/* The only remaining constraints are standalone ones and therefore don't need to be considered */
		return true;
		found:;
	}
	const constraint *cc = c[i];
	/* Swap it into the last slot if needed */
	if(i != --num)
	{
		c[i] = c[num];
		c[num] = cc;
	}
	return restrict_recurse_beta(fixed,mask,c,num,cc);
}

bool contains_encoding(const opcode_subset *o,const encoding *e)
{
	/* Early rejection checks - fixed bits must match, o musn't contain a nomatch constraint that's within e */
	if((o->fixed & e->mask) != (e->fixed & o->mask))
		return false;
	for(size_t i=0;i<o->nomatch.size();i++)
		for(uint32_t j=0;j<e->numconstraints;j++)
			if(o->nomatch[i] == e->constraints[j])
				return false;
	/* Work out the fixed bits & mask that any matching opcodes must adhere to */
	uint32_t fixed = o->fixed | e->fixed;
	uint32_t mask = o->mask | e->mask;
	/* Build a big list of constraints that must be satisfied */
	uint32_t num = o->match.size()+o->nomatch.size()+e->numconstraints;
	const constraint *c[num],*c2[o->nomatch.size()];
	uint32_t i=0;
	/* Invert the nomatch constraints for convenience. Although the mallocs may slow things down more than they help? */
	/* TODO - detect and split 'in' constraints */
	/* TODO - ignore all constraints in 'e' that don't touch any constraints in 'o', and vice-versa */
	for(size_t j=0;j<o->nomatch.size();j++)
	{
		constraint **p = new constraint *[1];
		p[0] = (constraint *) o->nomatch[j];
		assert(p[0]);
		p[0]->reference();
		c2[j] = c[i++] = new constraint_func(funcdef_lnot,p);
		assert(c2[j]);
	}
	for(size_t j=0;j<o->match.size();j++)
		c[i++] = o->match[j];
	for(uint32_t j=0;j<e->numconstraints;j++)
		c[i++] = e->constraints[j];
	assert(i==num);
	/* Now recursively restrict until we succeed or fail */
	bool ret = restrict_recurse(fixed,mask,c,num);
	for(size_t j=0;j<o->nomatch.size();j++)
		constraint::dereference((constraint *) c2[j]);
	return ret;
}

node *determine_best_node(const opcode_subset *o,const encodingset &e)
{
	if(readcache && badcache)
		return NULL;
	/* If the set of kin contains only one encoding, this is a leaf node */
	encodingset *kin = compute_kin(e);
	if(kin->size() == 1)
	{
		if(readcache)
		{
			/* Validate that the cache thinks there should be a leaf node here. If not, the cache must be out of date. */
			char c;
			if(!cache_read(&c,sizeof(c)))
				return NULL;
			if(c != 'l')
			{
				fprintf((updatecache?stdout:stderr),"Cache out of date\n");
				badcache = true;
				return NULL;
			}
		}
		else if(writecache)
		{
			char c = 'l';
			cache_write(&c,sizeof(c));
		}
		delete kin;
		return new node_leaf(e[0]);
	}
	if(verbose)
	{
		printf("determine_best_node: fixed %x mask %x with +%d -%d constraints for %d encodings (%d kin)\n",o->fixed,o->mask,(int)o->match.size(),(int)o->nomatch.size(),(int)e.size(),(int)kin->size());
		if(o->match.size())
		{
			printf("  ");
			for(size_t i=0;i<o->match.size();i++)
			{
				string c = o->match[i]->describe(NULL,0);
				printf(" +{%s}",c.c_str());
			}
			printf("\n");
		}
		if(o->nomatch.size())
		{
			printf("  ");
			for(size_t i=0;i<o->nomatch.size();i++)
			{
				string c = o->nomatch[i]->describe(NULL,0);
				printf(" -{%s}",c.c_str());
			}
			printf("\n");
		}
	}
	/* Store the parameters of the best location, rather than create nodes as we go. This should help speed things up. */
	const constraint *bestc=NULL;
	uint8_t beststartbit=0;
	uint8_t bestbitcount=0;
	encodingset *bestes=NULL;
	ntype besttype=ELEAF; /* fudgey uninitialised value */
	infogain bestgain = -1.0f;
	if(readcache)
	{
		char c;
		if(!cache_read(&c,sizeof(c)))
			return NULL;
		if(c == 'c')
		{
			/* Constraint names can bloat the cache file quite a bit, so we cache previously used constraints and just refer to them by an id instead */
			size_t id;
			if(!cache_read(&id,sizeof(id)))
				return NULL;
			if(id == cacheconstraints.size())
			{
				size_t len;
				if(!cache_read(&len,sizeof(len)))
					return NULL;
				char *buf = (char *) malloc(len+1);
				if(!buf)
				{
					fprintf((updatecache?stdout:stderr),"Failed to allocate string buffer for constraint name - cache corrupt?\n");
					badcache = true;
					return NULL;
				}
				if(!cache_read(buf,len))
				{
					delete[] buf;
					return NULL;
				}
				buf[len] = 0;
				/* Find this constraint */
				for(size_t i=0;i<e.size();i++)
				{
					const encoding *ee = e[i];
					if(ee->fields)
					{
						for(uint32_t j=0;j<ee->numconstraints;j++)
						{
							const constraint *c = ee->constraints[j];
							string name = c->describe(NULL,0);
							if(!name.compare(buf))
							{
								delete[] buf;
								cacheconstraints.push_back(c);
								goto found;
							}
						}
					}
				}
				fprintf((updatecache?stdout:stderr),"Failed to find constraint listed in cache file - cache file corrupt or out of date\n");
				delete[] buf;
				badcache = true;
				return NULL;
			}
			else if(id > cacheconstraints.size())
			{
				fprintf((updatecache?stdout:stderr),"Cache file corrupt or out of date\n");
				badcache = true;
				return NULL;
			}
found:
			bestes = split_on_constraint(o,e,cacheconstraints[id]);
			bestc = cacheconstraints[id];
			besttype = ECONSTRAINT;
			bestgain = compute_gain(kin,bestes,2);
			goto done;
		}
		else if(c == 'm')
		{
			if(!cache_read(&beststartbit,sizeof(beststartbit)))
				return NULL;
			if(!cache_read(&bestbitcount,sizeof(bestbitcount)))
				return NULL;
			if((beststartbit+bestbitcount > opcodesize) || !bestbitcount)
			{
				fprintf((updatecache?stdout:stderr),"Cache file corrupt or out of date\n");
				badcache = true;
				return NULL;
			}
			bestes = split_on_field(o,e,beststartbit,bestbitcount);
			besttype = EMASK;
			bestgain = compute_gain(kin,bestes,1<<bestbitcount);
			goto done;
		}
		else
		{
			fprintf((updatecache?stdout:stderr),"Cache file corrupt or out of date\n");
			badcache = true;
			return NULL;
		} 
	}
	else
	{
		/* First evaluate the worth of the constraints */
		constraintset attempted = o->match;
		for(size_t i=0;i<o->nomatch.size();i++)
			attempted.push_back(o->nomatch[i]);
		uint32_t search = 0;
		for(size_t i=0;i<e.size();i++)
		{
			const encoding *ee = e[i];
			/* If it has constraints but no fields, it's the fugly machine-generated encoding for all the undefined instructions. So skip it, since it takes ages to evaluate! */
			if(ee->fields)
			{
				for(uint32_t j=0;j<ee->numconstraints;j++)
				{
					const constraint *c = ee->constraints[j];
					size_t k;
					for(k=0;k<attempted.size();k++)
						if(attempted[k] == c)
							break;
					if(k == attempted.size())
					{
#ifdef EXTRAVERBOSE
						if(verbose) printf("-> evaluating constraint %u of encoding %s %s\n",j,ee->name,ee->text);
#endif
						attempted.push_back(c);
						encodingset *es = split_on_constraint(o,e,c);
						infogain gain = compute_gain(kin,es,2);
#ifdef EXTRAVERBOSE
						if(verbose) printf("   gain of %f\n",gain);
#endif
						if((besttype==ELEAF) || (gain > bestgain))
						{
							delete[] bestes;
							bestes = es;
							bestc = c;
							besttype = ECONSTRAINT;
							bestgain = gain;
							if(bestgain == 1.0f)
								goto done;
						}
						else
							delete[] es;
					}
				}
			}
			/* Compute search area for bitfield evaluation */
			search |= ee->getfields();
		}
		attempted.clear();
		/* Now evaluate the worth of arbitrary bitfields */
		/* We restrict the search area to bits that haven't been decoded yet (duh!), and bits for which a constraint exists in at least one encoding */
		/* Bigger masks will always result in a gain that's greater than or equal to a smaller mask, so we only need to worry about testing the largest area that fits in each position */
		search &= ~o->mask; /* Clear the decoded bits from the search area */
		for(uint32_t startbit=0;startbit<opcodesize;)
		{
			/* Work out the number of contiguous bits at this location */
			uint32_t bitcount=0;
			uint32_t size=1;
			uint32_t mask=0;
			while((search & (size<<startbit)) && (bitcount < maxmaskbits))
			{
				bitcount++;
				mask |= size<<startbit;
				size <<= 1;
			}
			if(!mask)
			{
				startbit++;
				continue;
			}
			/* Now generate nodes at startbit, startbit+1, etc. until we reach a point where moving +1 would reduce the search window size */
			do {
				encodingset *es = split_on_field(o,e,startbit,bitcount);
				infogain gain = compute_gain(kin,es,size);
				if((besttype==ELEAF) || (gain > bestgain))
				{
					delete[] bestes;
					bestes = es;
					beststartbit = startbit;
					bestbitcount = bitcount;
					besttype = EMASK;
					bestgain = gain;
					if(bestgain == 1.0f)
						goto done;
				}
				else
					delete[] es;
				if(!(search & (size<<startbit)))
					break;
				startbit++;
				mask <<= 1;
			} while(1);
			/* Skip the remaining bits */
			startbit += bitcount+1; /* bitcount+1 since we know bit(startbit+bitcount) is zero */
		}
	}
done:
	switch(besttype)
	{
	case ECONSTRAINT:
		if(verbose)
		{
			string c = bestc->describe(NULL,0);
			printf("-> best was constraint %s with infogain of %f\n",c.c_str(),bestgain);
		}
		if(writecache)
		{
			char c = 'c';
			cache_write(&c,sizeof(c));
			string name = bestc->describe(NULL,0);
			size_t id;
			for(id=0;id<cacheconstraints.size();id++)
			{
				if((cacheconstraints[id] == bestc) || !cacheconstraints[id]->describe(NULL,0).compare(name))
					break;
			}
			cache_write(&id,sizeof(id));
			if(id == cacheconstraints.size())
			{
				cacheconstraints.push_back(bestc);
				size_t len = name.size();
				cache_write(&len,sizeof(len));
				cache_write(name.c_str(),len);
			}
		}
		return new node_constraint(bestc,o,bestes);
	case EMASK:
		if(verbose) printf("-> best was mask of %x with infogain of %f\n",((((uint32_t)1)<<bestbitcount)-1)<<beststartbit,bestgain);
		if(writecache)
		{
			char c = 'm';
			cache_write(&c,sizeof(c));
			cache_write(&beststartbit,sizeof(beststartbit));
			cache_write(&bestbitcount,sizeof(bestbitcount));
		}
		return new node_mask(beststartbit,bestbitcount,o,bestes);
	default:
		assert(0);
		return NULL;
	}
}

encodingset *split_on_constraint(const opcode_subset *o,const encodingset &e,const constraint *c)
{
	encodingset *es = new encodingset[2];
	es[0].reserve(e.size()); /* Assume worst-case to avoid reallocs */
	es[1].reserve(e.size());
	opcode_subset yes(*o),no(*o);
	yes.match.push_back(c);
	no.nomatch.push_back(c);
	for(size_t i=0;i<e.size();i++)
	{
		if(contains_encoding(&yes,e[i]))
			es[0].push_back(e[i]);
		if(contains_encoding(&no,e[i]))
			es[1].push_back(e[i]);
	}
	return es;
}

encodingset *split_on_field_fast(const opcode_subset *o,const encodingset &e,uint32_t startbit,uint32_t bitcount)
{
	/* Faster version of split_on_field, only works if there are no opcode match/nomatch constraints
	   Note that even though split_on_field has been rewritten and made faster since this function was written, this function still seems to be faster when used for completeness checks */
	assert((o->match.size() == 0) && (o->nomatch.size() == 0));
	uint32_t size = ((uint32_t)1)<<bitcount;
	encodingset *es = new encodingset[size];
	uint32_t bits = (size-1)<<startbit;
	uint32_t mask = bits | o->mask;
	for(uint32_t f=0;f<size;f++)
	{
		uint32_t fixed = (f<<startbit) | o->fixed;
		es[f].reserve(e.size());
		for(size_t i=0;i<e.size();i++)
		{
			uint32_t fields = e[i]->getfields();
			if((fields & mask) == fields)
			{
				/* Result will be const for entire subset */
				if(e[i]->matches(fixed))
					es[f].push_back(e[i]);
			}
			else if((bits & fields)==0)
			{
				/* Field doesn't touch any constrained bits, so if encoding was in parent it'll definitely be here too */
				es[f].push_back(e[i]);
			}
			else if((fixed & e[i]->mask) == (e[i]->fixed & mask)) /* Is a match even possible? (does the subset match the fixed bits) */
			{
				/* Early rejection check; look for any constraints whose fields are fully within 'mask' and check if they fail */
				/* Also an early acceptance check: if there aren't any constraints which are half inside and half outside 'mask' (i.e. they only lie fully inside or fully outside), then by evaluating the ones that lie inside we know whether this encoding fits in the subset */
				bool reject = false;
				bool fits = true;
				for(uint32_t j=0;j<e[i]->numconstraints;j++)
				{
					uint32_t cfields = e[i]->constraints[j]->getfields();
					
					if(((cfields & mask) == cfields) && !e[i]->constraints[j]->eval(fixed))
					{
						reject = true;
						break;
					}
					if(((cfields & mask) != cfields) && ((cfields & mask) != 0))
						fits = false;
				}
				if(reject)
					continue;
				if(fits)
				{
					es[f].push_back(e[i]);
					continue;
				}
				/* Must iterate */
				/* TODO - further optimisation? - we only need to check the constraints that intersect 'bits' */
				uint32_t fixed2 = fixed | e[i]->fixed; /* No point checking opcodes which won't match! */
				uint32_t mask2 = mask | e[i]->mask | (opcodebits & ~e[i]->getfields()); /* Mark the fixed bits and the non-constrained fields as fixed - they won't affect our evaluation */
				uint32_t opcode = 0;
				do {
					if(e[i]->matches(opcode | fixed2))
					{
						es[f].push_back(e[i]);
						break;
					}
					opcode++;
					while(opcode & mask2)
						opcode = (opcode+mask2)&~mask2;
				} while(opcode & opcodebits);
			}
		}
	}
	return es;
}

encodingset *split_on_field(const opcode_subset *o,const encodingset &e,uint32_t startbit,uint32_t bitcount)
{
	uint32_t size = ((uint32_t)1)<<bitcount;
	encodingset *es = new encodingset[size];
	uint32_t bits = (size-1)<<startbit;
	opcode_subset o2(*o);
	o2.mask = bits | o->mask;
	for(uint32_t f=0;f<size;f++)
	{
		es[f].reserve(e.size());
		o2.fixed = (f<<startbit) | o->fixed;
		for(size_t i=0;i<e.size();i++)
		{
			if(contains_encoding(&o2,e[i]))
			{
				es[f].push_back(e[i]);
			}
		}
	}
	return es;
}

#if 0
static bool kinsets_match(const encodingset *a,const encodingset *b)
{
	if(a->size() != b->size())
		return false;
	for(size_t i=0;i<a->size();i++)
	{
		const encoding *ea = (*a)[i];
		size_t j;
		for(j=0;j<b->size();j++)
			if(ea == (*b)[j])
				break;
		if(j == b->size())
			return false;
	}
	return true;
}
#endif

infogain compute_gain(const encodingset *kin,const encodingset *es,uint32_t count)
{
	/* Basic measurement of information gain:
	   convert each es element into kin list
	   compute total kin count, using a count of 0 if kinset only contains one element (i.e. fully classified)
	   return 1-(total/(kin->size()*count))
	   i.e. 1.0 if everything is classified, 0.0 if no gain from current state
	   This isn't the correct way of calculating entropy/gain in ID3, but it seems to result in much better (i.e. shorter, smaller) trees than my experiments with more proper algorithms */
	uint32_t total=0;
	encodingset solvedset;
	uint32_t worstcase = kin->size()*count;
	encodingset *kinsets[count];
	memset(kinsets,0,sizeof(encodingset *)*count);
	for(uint32_t i=0;i<count;i++)
	{
		/* Fuzzy trees can contain dead-end paths
		   Regular trees may generate them too (e.g. if only two encodings are left, it tests a constraint, and the constraint doesn't help)
		   Or in a more general case, if it tests on a field and a match/nomatch constraint prevents all the children from being populated */
		if(es[i].size() == 0)
		{
			worstcase -= kin->size(); /* Dead end paths should be ignored in gain calculations */
			continue;
		}
		encodingset *kinset = kinsets[i] = compute_kin(es[i]);
#if 0
		if(verbose)
		{
			printf("   [%u] =",i);
			for(size_t j=0;j<kinset->size();j++)
				printf(" %s",(*kinset)[j]->name);
			printf("\n");
		}
#endif
#if 0
		/* Ignore childsets which are identical to previous ones (i.e. duplicated child nodes) */
		for(uint32_t k=0;k<i;k++)
		{
			if(kinsets[k] && kinsets_match(kinsets[k],kinset))
			{
				worstcase -= kin->size()-2; /* Slight weighting to ensure the result is at least partially penalised */
				delete kinsets[i];
				kinsets[i] = NULL;
				break;
			}
		}
		if(!kinsets[i])
			continue;
#endif
#if 1
		if(kinset->size() > 1)
			total += kinset->size();
		else
		{
			/* Penalise duplicated solutions */
			size_t j;
			for(j=0;j<solvedset.size();j++)
				if(solvedset[j] == (*kinset)[0])
					break;
			if(j == solvedset.size())
				solvedset.push_back((*kinset)[0]);
			else
				total += 1;
		}
#else
		for(size_t k=0;k<kinset->size();k++)
		{
			/* Penalise duplicated solutions */
			size_t j;
			for(j=0;j<solvedset.size();j++)
				if(solvedset[j] == (*kinset)[k])
					break;
			if(j == solvedset.size())
				solvedset.push_back((*kinset)[k]);
			else
				total += 1;
		}
#endif
	}
	for(uint32_t i=0;i<count;i++)
		delete kinsets[i];
	total = worstcase-total;
	return ((float)total)/((float)worstcase);
}
